@@ -23,36 +23,90 @@ import os
 import re
 import base64
 import tempfile
+import shlex
+import shutil
+import subprocess
+import sys
 from collections import defaultdict, namedtuple
 from functools import partial
 
-from reportlab.pdfgen.pdfimages import PDFImage
-from reportlab.pdfbase.pdfmetrics import stringWidth
+from reportlab.pdfbase.pdfmetrics import registerFont, stringWidth
+from reportlab.pdfbase.ttfonts import TTFError, TTFont
 from reportlab.pdfgen.canvas import FILL_EVEN_ODD, FILL_NON_ZERO
+from reportlab.pdfgen.pdfimages import PDFImage
 from reportlab.graphics.shapes import (
     _CLOSEPATH, Circle, Drawing, Ellipse, Group, Image, Line, Path, PolyLine,
     Polygon, Rect, String,
 )
 from reportlab.lib import colors
 from reportlab.lib.units import pica, toLength
+from reportlab.lib.utils import haveImages
 from lxml import etree
+import cssselect2
+import tinycss2
 
 from .utils import (
     bezier_arc_from_end_points, convert_quadratic_to_cubic_path,
     normalise_svg_path,
 )
 
-__version__ = '0.8.1'
+__version__ = '0.9.0'
 __license__ = 'LGPL 3'
 __author__ = 'Dinu Gherman'
-__date__ = '2017-04-22'
+__date__ = '2018-12-08'
 
 XML_NS = 'http://www.w3.org/XML/1998/namespace'
+STANDARD_FONT_NAMES = (
+    'Times-Roman', 'Times-Italic', 'Times-Bold', 'Times-BoldItalic',
+    'Helvetica', 'Helvetica-Oblique', 'Helvetica-Bold', 'Helvetica-BoldOblique',
+    'Courier', 'Courier-Oblique', 'Courier-Bold', 'Courier-BoldOblique',
+    'Symbol', 'ZapfDingbats',
+)
+DEFAULT_FONT_NAME = "Helvetica"
+_registered_fonts = {}
 
 logger = logging.getLogger(__name__)
 
 Box = namedtuple('Box', ['x', 'y', 'width', 'height'])
 
+split_whitespace = re.compile('[^ \t\r\n\f]+').findall
+
+
+def find_font(font_name):
+    """Return the font and a Boolean indicating if the match is exact."""
+    if font_name in STANDARD_FONT_NAMES:
+        return font_name, True
+    elif font_name in _registered_fonts:
+        return font_name, _registered_fonts[font_name]
+
+    NOT_FOUND = (None, False)
+    try:
+        # Try first to register the font if it exists as ttf,
+        # based on ReportLab font search.
+        registerFont(TTFont(font_name, '%s.ttf' % font_name))
+        _registered_fonts[font_name] = True
+        return font_name, True
+    except TTFError:
+        # Try searching with Fontconfig
+        try:
+            pipe = subprocess.Popen(
+                ['fc-match', '-s', '--format=%{file}\\n', font_name],
+                stdout=subprocess.PIPE,
+                stderr=subprocess.PIPE,
+            )
+            output = pipe.communicate()[0].decode(sys.getfilesystemencoding())
+            font_path = output.split('\n')[0]
+        except OSError:
+            return NOT_FOUND
+        try:
+            registerFont(TTFont(font_name, font_path))
+        except TTFError:
+            return NOT_FOUND
+        # Fontconfig may return a default font totally unrelated with font_name
+        exact = font_name.lower() in os.path.basename(font_path).lower()
+        _registered_fonts[font_name] = exact
+        return font_name, exact
+
 
 class NoStrokePath(Path):
     """
@@ -90,11 +144,35 @@ class ClippingPath(Path):
         return props
 
 
+class CSSMatcher(cssselect2.Matcher):
+    def __init__(self, style_content):
+        super(CSSMatcher, self).__init__()
+        self.rules = tinycss2.parse_stylesheet(
+            style_content, skip_comments=True, skip_whitespace=True
+        )
+        for rule in self.rules:
+            if not rule.prelude:
+                continue
+            selectors = cssselect2.compile_selector_list(rule.prelude)
+            selector_string = tinycss2.serialize(rule.prelude)
+            content_dict = dict(
+                (attr.split(':')[0].strip(), attr.split(':')[1].strip())
+                for attr in tinycss2.serialize(rule.content).split(';')
+                if ':' in attr
+            )
+            payload = (selector_string, content_dict)
+            for selector in selectors:
+                self.add_selector(selector, payload)
+
+
 # Attribute converters (from SVG to RLG)
 
-class AttributeConverter:
+class AttributeConverter(object):
     "An abstract class to locate and convert attributes in a DOM instance."
 
+    def __init__(self):
+        self.css_rules = None
+
     def parseMultiAttributes(self, line):
         """Try parsing compound attribute string.
 
@@ -123,6 +201,11 @@ class AttributeConverter:
 
         # This needs also to lookup values like "url(#SomeName)"...
 
+        if self.css_rules is not None and not svgNode.attrib.get('__rules_applied', False):
+            if isinstance(svgNode, NodeTracker):
+                svgNode.apply_rules(self.css_rules)
+            else:
+                ElementWrapper(svgNode).apply_rules(self.css_rules)
         attr_value = svgNode.attrib.get(name, '').strip()
 
         if attr_value and attr_value != "inherit":
@@ -131,9 +214,8 @@ class AttributeConverter:
             dict = self.parseMultiAttributes(svgNode.attrib.get("style"))
             if name in dict:
                 return dict[name]
-        elif svgNode.getparent() is not None:
+        if svgNode.getparent() is not None:
             return self.findAttr(svgNode.getparent(), name)
-
         return ''
 
     def getAllAttributes(self, svgNode):
@@ -180,10 +262,13 @@ class AttributeConverter:
             subline = subline.strip()
             subline = subline.replace(',', ' ')
             subline = re.sub("[ ]+", ',', subline)
-            if ',' in subline:
-                indices.append(tuple(float(num) for num in subline.split(',')))
-            else:
-                indices.append(float(subline))
+            try:
+                if ',' in subline:
+                    indices.append(tuple(float(num) for num in subline.split(',')))
+                else:
+                    indices.append(float(subline))
+            except ValueError:
+                continue
             ops = ops[:bi] + ' '*(bj-bi+1) + ops[bj+1:]
         ops = ops.replace(',', ' ').split()
 
@@ -201,13 +286,18 @@ class AttributeConverter:
 class Svg2RlgAttributeConverter(AttributeConverter):
     "A concrete SVG to RLG attribute converter."
 
-    def __init__(self,color_converter=None):
+    def __init__(self, color_converter=None):
+        super(Svg2RlgAttributeConverter, self).__init__()
         self.color_converter = color_converter or self.identity_color_converter
 
     @staticmethod
     def identity_color_converter(c):
         return c
 
+    @staticmethod
+    def split_attr_list(attr):
+        return shlex.split(attr.strip().replace(',', ' '))
+
     def convertLength(self, svgAttr, percentOf=100, em_base=12):
         "Convert length to points."
 
@@ -235,19 +325,13 @@ class Svg2RlgAttributeConverter(AttributeConverter):
             text = text.replace("ex", '')
 
         text = text.strip()
-        length = toLength(text)
+        length = toLength(text)  # this does the default measurements such as mm and cm
 
         return length
 
     def convertLengthList(self, svgAttr):
-        "Convert a list of lengths."
-
-        t = svgAttr.replace(',', ' ')
-        t = t.strip()
-        t = re.sub("[ ]+", ' ', t)
-        a = t.split(' ')
-
-        return [self.convertLength(a) for a in a]
+        """Convert a list of lengths."""
+        return [self.convertLength(a) for a in self.split_attr_list(svgAttr)]
 
     def convertOpacity(self, svgAttr):
         return float(svgAttr)
@@ -288,7 +372,7 @@ class Svg2RlgAttributeConverter(AttributeConverter):
             return self.color_converter(colors.HexColor(col))
         elif text.startswith('rgb') and '%' in text:
             t = text[3:].replace('%', '').strip('()')
-            tup = (int(val)/100.0 for val in t.split(','))
+            tup = (float(val)/100.0 for val in t.split(','))
             return self.color_converter(colors.Color(*tup))
 
         logger.warn("Can't handle color: %s" % text)
@@ -310,34 +394,99 @@ class Svg2RlgAttributeConverter(AttributeConverter):
         return strokeDashOffset
 
     def convertFontFamily(self, svgAttr):
+        if not svgAttr:
+            return ''
         # very hackish
         font_mapping = {
-            "sans-serif":"Helvetica",
-            "serif":"Times-Roman",
-            "monospace":"Courier"
+            "sans-serif": "Helvetica",
+            "serif": "Times-Roman",
+            "times": "Times-Roman",
+            "monospace": "Courier",
         }
-        font_name = svgAttr
-        if not font_name:
-            return ''
-        try:
-            font_name = font_mapping[font_name]
-        except KeyError:
-            pass
-        if font_name not in ("Helvetica", "Times-Roman", "Courier"):
-            font_name = "Helvetica"
-
-        return font_name
+        font_names = [
+            font_mapping.get(font_name.lower(), font_name)
+            for font_name in self.split_attr_list(svgAttr)
+        ]
+        non_exact_matches = []
+        for font_name in font_names:
+            font_name, exact = find_font(font_name)
+            if exact:
+                return font_name
+            elif font_name:
+                non_exact_matches.append(font_name)
+        if non_exact_matches:
+            return non_exact_matches[0]
+        else:
+            logger.warn("Unable to find a suitable font for 'font-family:%s'" % svgAttr)
+            return DEFAULT_FONT_NAME
 
 
-class NodeTracker:
+class ElementWrapper(object):
+    """
+    lxml element wrapper to partially match the API from cssselect2.ElementWrapper
+    so as element can be passed to rules.match().
+    """
+    in_html_document = False
+
+    def __init__(self, obj):
+        self.object = obj
+
+    @property
+    def id(self):
+        return self.object.attrib.get('id')
+
+    @property
+    def etree_element(self):
+        return self.object
+
+    @property
+    def parent(self):
+        par = self.object.getparent()
+        return ElementWrapper(par) if par else None
+
+    @property
+    def classes(self):
+        cl = self.object.attrib.get('class')
+        return split_whitespace(cl) if cl is not None else []
+
+    @property
+    def local_name(self):
+        return node_name(self.object)
+
+    @property
+    def namespace_url(self):
+        if '}' in self.object.tag:
+            self.object.tag.split('}')[0][1:]
+
+    def iter_ancestors(self):
+        element = self
+        while element.parent is not None:
+            element = element.parent
+            yield element
+
+    def apply_rules(self, rules):
+        matches = rules.match(self)
+        for match in matches:
+            attr_dict = match[3][1]
+            for attr, val in attr_dict.items():
+                if not attr in self.object.attrib:
+                    try:
+                        self.object.attrib[attr] = val
+                    except ValueError:
+                        pass
+        # Set marker on the node to not apply rules more than once
+        self.object.set('__rules_applied', '1')
+
+
+class NodeTracker(ElementWrapper):
     """An object wrapper keeping track of arguments to certain method calls.
 
     Instances wrap an object and store all arguments to one special
     method, getAttribute(name), in a list of unique elements, usedAttrs.
     """
 
-    def __init__(self, anObject):
-        self.object = anObject
+    def __init__(self, obj):
+        super(NodeTracker, self).__init__(obj)
         self.usedAttrs = []
 
     def getAttribute(self, name):
@@ -367,16 +516,17 @@ class SvgRenderer:
         self.handled_shapes = self.shape_converter.get_handled_shapes()
         self.definitions = {}
         self.waiting_use_nodes = defaultdict(list)
-        self.box = Box(x=0, y=0, width=0, height=0)
 
     def render(self, svg_node):
-        main_group = self.renderNode(svg_node)
+        node = NodeTracker(svg_node)
+        main_group = self.renderSvg(node, outermost=True)
         for xlink in self.waiting_use_nodes.keys():
             logger.debug("Ignoring unavailable object width ID '%s'." % xlink)
 
+        view_box = self.get_box(node, default_box=True)
         main_group.scale(1, -1)
-        main_group.translate(0 - self.box.x, -self.box.height - self.box.y)
-        drawing = Drawing(self.box.width, self.box.height)
+        main_group.translate(0 - view_box.x, -view_box.height - view_box.y)
+        drawing = Drawing(view_box.width, view_box.height)
         drawing.add(main_group)
         return drawing
 
@@ -389,10 +539,7 @@ class SvgRenderer:
 
         clipping = self.get_clippath(n)
         if name == "svg":
-            if n.getAttribute("{%s}space" % XML_NS) == 'preserve':
-                self.shape_converter.preserve_space = True
-            item = self.renderSvg(n)
-            return item
+            return self.renderSvg(n)
         elif name == "defs":
             item = self.renderG(n)
         elif name == 'a':
@@ -403,6 +550,8 @@ class SvgRenderer:
             item = self.renderG(n, clipping=clipping)
             if display != "none":
                 parent.add(item)
+        elif name == "style":
+            self.renderStyle(n)
         elif name == "symbol":
             item = self.renderSymbol(n)
             parent.add(item)
@@ -469,19 +618,43 @@ class SvgRenderer:
         # Main SVG desc. attr. could be used in the PDF document info field.
         pass
 
-    def renderSvg(self, node):
+    def get_box(self, svg_node, default_box=False):
+        view_box = svg_node.getAttribute("viewBox")
+        if view_box:
+            view_box = self.attrConverter.convertLengthList(view_box)
+            return Box(*view_box)
+        if default_box:
+            width, height = map(svg_node.getAttribute, ("width", "height"))
+            width, height = map(self.attrConverter.convertLength, (width, height))
+            return Box(0, 0, width, height)
+
+    def renderSvg(self, node, outermost=False):
         getAttr = node.getAttribute
-        width, height = map(getAttr, ("width", "height"))
-        width, height = map(self.attrConverter.convertLength, (width, height))
-        viewBox = getAttr("viewBox")
-        if viewBox:
-            viewBox = self.attrConverter.convertLengthList(viewBox)
-            self.box = Box(*viewBox)
-        else:
-            self.box = Box(0, 0, width, height)
+
+        _saved_preserve_space = self.shape_converter.preserve_space
+        self.shape_converter.preserve_space = getAttr("{%s}space" % XML_NS) == 'preserve'
+
         group = Group()
         for child in node.getchildren():
             self.renderNode(child, group)
+        self.shape_converter.preserve_space = _saved_preserve_space
+
+        if not outermost:
+            x, y = map(getAttr, ("x", "y"))
+            x, y = map(self.attrConverter.convertLength, (x, y))
+            if x or y:
+                group.translate(x or 0, y or 0)
+
+            view_box = self.get_box(node)
+            if view_box:
+                x_scale, y_scale = 1, 1
+                width, height = map(getAttr, ("width", "height"))
+                width, height = map(self.attrConverter.convertLength, (width, height))
+                if view_box.height != height:
+                    y_scale = height / view_box.height
+                if view_box.width != width:
+                    x_scale = width / view_box.width
+                group.scale(x_scale, y_scale)
         return group
 
     def renderG(self, node, clipping=None, display=1):
@@ -500,6 +673,9 @@ class SvgRenderer:
 
         return gr
 
+    def renderStyle(self, node):
+        self.attrConverter.css_rules = CSSMatcher(node.text)
+
     def renderSymbol(self, node):
         return self.renderG(node, display=0)
 
@@ -686,7 +862,7 @@ class Svg2RlgShapeConverter(SvgShapeConverter):
 
         dx0, dy0 = 0, 0
         x1, y1 = 0, 0
-        ff = attrConv.findAttr(node, "font-family") or "Helvetica"
+        ff = attrConv.findAttr(node, "font-family") or DEFAULT_FONT_NAME
         ff = attrConv.convertFontFamily(ff)
         fs = attrConv.findAttr(node, "font-size") or "12"
         fs = attrConv.convertLength(fs)
@@ -744,7 +920,7 @@ class Svg2RlgShapeConverter(SvgShapeConverter):
         subpath_start = []
         lastop = ''
 
-        for i in xrange(0, len(normPath), 2):
+        for i in range(0, len(normPath), 2):
             op, nums = normPath[i:i+2]
 
             if op in ('m', 'M') and i > 0 and path.operators[-1] != _CLOSEPATH:
@@ -894,33 +1070,41 @@ class Svg2RlgShapeConverter(SvgShapeConverter):
         return gr
 
     def convertImage(self, node):
-        logger.warn("Adding box instead of image.")
+        if not haveImages:
+            logger.warn(
+                "Unable to handle embedded images. Maybe the pillow library is missing?"
+            )
+            return None
+
         getAttr = node.getAttribute
         x, y, width, height = map(getAttr, ('x', 'y', "width", "height"))
         x, y, width, height = map(self.attrConverter.convertLength, (x, y, width, height))
         xlink_href = node.attrib.get('{http://www.w3.org/1999/xlink}href')
 
-        magic = "data:image/jpeg;base64"
-        if xlink_href[:len(magic)] == magic:
-            pat = "data:image/(\w+?);base64"
-            ext = re.match(pat, magic).groups()[0]
-            jpeg_data = base64.decodestring(xlink_href[len(magic):].encode('ascii'))
-            _, path = tempfile.mkstemp(suffix='.%s' % ext)
+        magic_re = r"^data:image/(jpeg|png);base64"
+        match = re.match(magic_re, xlink_href)
+        if match:
+            img_format = match.groups()[0]
+            image_data = base64.decodestring(xlink_href[(match.span(0)[1] + 1):].encode('ascii'))
+            _, path = tempfile.mkstemp(suffix='.%s' % img_format)
             with open(path, 'wb') as fh:
-                fh.write(jpeg_data)
-            img = Image(int(x), int(y+height), int(width), int(-height), path)
+                fh.write(image_data)
+            img = Image(int(x), int(y + height), int(width), int(height), path)
             # this needs to be removed later, not here...
             # if exists(path): os.remove(path)
         else:
             xlink_href = os.path.join(os.path.dirname(self.svg_source_file), xlink_href)
-            img = Image(int(x), int(y+height), int(width), int(-height), xlink_href)
+            img = Image(int(x), int(y + height), int(width), int(height), xlink_href)
             try:
                 # this will catch invalid image
                 PDFImage(xlink_href, 0, 0)
             except IOError:
                 logger.error("Unable to read the image %s. Skipping..." % img.path)
                 return None
-        return img
+        group = Group(img)
+        group.translate(0, (x + height) * 2)
+        group.scale(1, -1)
+        return group
 
     def applyTransformOnGroup(self, transform, group):
         """Apply an SVG transformation to a RL Group shape.
@@ -982,7 +1166,7 @@ class Svg2RlgShapeConverter(SvgShapeConverter):
             ("stroke-dasharray", "strokeDashArray", "convertDashArray", "none"),
         )
         mappingF = (
-            ("font-family", "fontName", "convertFontFamily", "Helvetica"),
+            ("font-family", "fontName", "convertFontFamily", DEFAULT_FONT_NAME),
             ("font-size", "fontSize", "convertLength", "12"),
             ("text-anchor", "textAnchor", "id", "start"),
         )
@@ -1021,8 +1205,8 @@ def svg2rlg(path, **kwargs):
     # unzip .svgz file into .svg
     unzipped = False
     if isinstance(path, str) and os.path.splitext(path)[1].lower() == ".svgz":
-        data = gzip.GzipFile(path, "rb").read()
-        open(path[:-1], 'w').write(data)
+        with gzip.open(path, 'rb') as f_in, open(path[:-1], 'wb') as f_out:
+            shutil.copyfileobj(f_in, f_out)
         path = path[:-1]
         unzipped = True
 
